/*
    Copyright 2008 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; version 3 of the License.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file gpl-3.0.txt. If not, see
    <http://www.gnu.org/licenses/>.
*/

#include "emea.h"
#include "emea.moc"
#include "emea_p.h"
#include "emea_p.moc"

#include <QtCore/QDebug>
#include <QtCore/QTimer>

#include <QtGui/QApplication>
#include <QtGui/QMenuBar>
#include <QtGui/QMenu>
#include <QtGui/QAction>
#include <QtGui/QTreeView>
#include <QtGui/QFileDialog>
#include <QtGui/QScrollArea>
#include <QtGui/QCloseEvent>

#include <thread_template.h>
#include <stackediconwidget.h>
#include <osciwidget_v3.h>
#include <qtreeitemmodel.h>
#include <qtreeitemmodeldelegate.h>
#include <graphicalguiserver.h>
#include <resultviewer.h>

#include "valuelogger.h"
#include "valuemeter.h"

class EMEAWindow::Private
{
	public:
		Private( QMainWindow* p ) {
			stack = new StackedIconWidget( p );
			osci = new OsciWidgetV3( 0 );
			stack->addWidget( osci );
			viewer = new ResultViewer( 0 );
			QScrollArea* tmp = new QScrollArea( p );
			tmp->setObjectName( viewer->objectName() );
			tmp->setWindowIcon( viewer->windowIcon() );
			tmp->setWidget( viewer );
			tmp->setWidgetResizable( true );
			stack->addWidget( tmp );

			treemodel = new NSCore::QTreeItemModel( p );

			treeview = new QTreeView( 0 );
			treeview->setObjectName( "Chain" );
			treeview->setWindowIcon( QIcon( ":/icons/kde/cr22-action-view_tree.png" ) );
			stack->addWidget( treeview );
			treeview->setItemDelegate( new QTreeItemModelDelegate( new GraphicalGuiServer( p ), p ) );
			treeview->setModel( treemodel );
			treeview->expandToDepth( 0 );

			spectrumguicreator = new SpectrumGuiCreator( p );
			p->connect( spectrumguicreator, SIGNAL( newSpectrumGui( QWidget* ) ), p, SLOT( addWidgetToStack( QWidget* ) ) );

			logger = new ValueLogging( p );
			//guilogger = new GuiLogging( p );
			//stack->addWidget( guilogger );
			guimeter = new ValueMeters( p );
			stack->addWidget( guimeter );

			finishedWatcher = 0;
		}

		StackedIconWidget* stack;
		OsciWidgetV3* osci;
		ResultViewer* viewer;

		QTreeView* treeview;
		NSCore::QTreeItemModel* treemodel;

		QAction* aosciconnect;
		QAction* aresultviewerconnect;

		QList<NSCore::DataSetProducer*> producers;
		QList< QList<NSCore::DataSetChainMember*> > chain;

		SpectrumGuiCreator* spectrumguicreator;

		QStringList arguments;

		EMEAFinishedWatcher* finishedWatcher;

		QList<QThread*> threads;

		ValueLogging* logger;
		//GuiLogging* guilogger;
		ValueMeters* guimeter;

		void updateChainConnections() {
			qDebug() << "EMEAWindow::Private::updateChainConnections()";
			foreach( NSCore::DataSetProducer* producer, producers )
				producer->unregisterAllConsumers();
			foreach( QList<NSCore::DataSetChainMember*> list, chain )
				foreach( NSCore::DataSetProducer* producer, list )
					producer->unregisterAllConsumers();
			if ( chain.size() > 0 ) {
				foreach( NSCore::DataSetProducer* producer, producers )
					foreach( NSCore::DataSetChainMember* consumer, chain.at( 0 ) )
						producer->registerConsumer( consumer );
			}
			for ( int i=1; i<chain.size(); ++i ) {
				foreach( NSCore::DataSetChainMember* producer, chain.at( i-1 ) ) {
					foreach( NSCore::DataSetChainMember* consumer, chain.at( i ) ) {
						qDebug() << i << "\t connecting:" << producer << consumer;
						producer->registerConsumer( consumer );
					}
				}
			}
			//if ( finishedWatcher )
			//	toggleTool( finishedWatcher, true );
		}
		QList<NSCore::QTreeItem*> getAllRootItems() {
			QList<NSCore::QTreeItem*> items;
			foreach( NSCore::DataSetProducer* producer, producers )
				items.append( producer );
			foreach( QList<NSCore::DataSetChainMember*> level, chain )
				foreach( NSCore::DataSetChainMember* member, level )
					items.append( member );
			return items;
		}

		void toggleTool( NSCore::DataSetChainMember* item, bool n ) {
			if ( ! chain.isEmpty() )
				foreach( NSCore::DataSetChainMember* member, chain.last() ) {
					member->unregisterConsumer( item );
					// Register at last members in chain
					if ( n && chain.size() > 0 )
						member->registerConsumer( item );
				}
			if ( ! producers.isEmpty() )
				foreach( NSCore::DataSetProducer* producer, producers ) {
					producer->unregisterConsumer( item );
					// Register at the producers directly
					if ( n && chain.size() == 0 )
						producer->registerConsumer( item );
				}
		}
};

EMEAWindow::EMEAWindow( QWidget* p )
	: QMainWindow( p )
	, d( new EMEAWindow::Private( this ) )
{
	qDebug() << "EMEAWindow::EMEAWindow(" << p << ")";

	setCentralWidget( d->stack );

	QMenu* filemenu = menuBar()->addMenu( "File" );
	QAction* aopen = new QAction( QIcon( ":/icons/kde/cr22-action-fileopen.png" ), "Open xml-file", this );
	connect( aopen, SIGNAL( triggered() ), this, SLOT( loadXmlFromFile() ) );
	filemenu->addAction( aopen );
	QAction* asave = new QAction( QIcon( ":/icons/kde/cr22-action-filesaveas.png" ), "Save to xml-file", this );
	connect( asave, SIGNAL( triggered() ), this, SLOT( saveXmlToFile() ) );
	filemenu->addAction( asave );
	filemenu->addSeparator();
	QAction* aquit = new QAction( QIcon( ":icons/kde/cr22-action-exit.png" ), "Quit", this );
	aquit->setShortcut( Qt::CTRL|Qt::Key_Q );
	connect( aquit, SIGNAL( triggered() ), this, SLOT( close() ) );
	filemenu->addAction( aquit );

	QMenu* toolsmenu = menuBar()->addMenu( "Tools" );
	d->aosciconnect = new QAction( "Connect Osci", this );
	d->aosciconnect->setCheckable( true );
	connect( d->aosciconnect, SIGNAL( toggled( bool ) ), this, SLOT( toggleOsci( bool ) ) );
	toolsmenu->addAction( d->aosciconnect );
	d->aresultviewerconnect = new QAction( "Connect Result-Viewer", this );
	d->aresultviewerconnect->setCheckable( true );
	connect( d->aresultviewerconnect, SIGNAL( toggled( bool ) ), this, SLOT( toggleViewer( bool ) ) );
	toolsmenu->addAction( d->aresultviewerconnect );

	d->arguments = QApplication::arguments();

	//connect( d->finishedWatcher, SIGNAL( evaluationFinished() ), this, SLOT( close() ) );

	QTimer::singleShot( 1, this, SLOT( lateInit() ) );
}
EMEAWindow::~EMEAWindow() {
	qDebug() << "EMEAWindow::~EMEAWindow()";
	while ( ! d->threads.isEmpty() )
		delete d->threads.takeFirst();
}

void EMEAWindow::closeEvent( QCloseEvent* ev ) {
	//qDebug() << "EMEAWindow::closeEvent(" << ev << ")";
	d->stack->unfloatAllWidgets();
	ev->accept();
}

void EMEAWindow::lateInit() {
	qDebug() << "EMEAWindow::lateInit()";
	/// Get the number of threads wanted/possible
	int threads = QThread::idealThreadCount();
	for ( QStringList::iterator it=d->arguments.begin(); it!=d->arguments.end(); ++it )
		if ( *it == "-threads" )
			threads = ( ++it )->toInt();
	threads = qMax( 1, threads );
	/// Set up the chain
	setupChain( threads );
	d->updateChainConnections();
	d->treeview->expandToDepth( 0 );

	/// React to the arguments given on the commandline
	// Load xml-files
	foreach( QString arg, d->arguments )
		if ( arg.endsWith( ".xml" ) )
			if ( QFile::exists( arg ) )
				loadXmlFromFile( arg );
	// Autostart the evaluation
	if ( d->arguments.contains( "-autostart" ) )
		QTimer::singleShot( 2000, this, SLOT( delayedAutoStart() ) );
	else {
		d->aosciconnect->setChecked( true );
		d->aresultviewerconnect->setChecked( true );
	}
	// Autostop the evaluation
	if ( d->arguments.contains( "-autostop" ) ) {
		qDebug() << "  Will initialize the FinishedWatcher...";
		d->finishedWatcher = new EMEAFinishedWatcher( this );
		d->finishedWatcher->start();
		connect( d->finishedWatcher, SIGNAL( evaluationFinished() ), this, SLOT( close() ) );
		d->toggleTool( d->finishedWatcher, true );
		qDebug() << "   is" << d->finishedWatcher;
	}
	qDebug() << " lateInit() finished";
}

void EMEAWindow::toggleOsci( bool n ) {
	qDebug() << "EMEAWindow::toggleOsci(" << n << ")";
	d->toggleTool( d->osci->chainMember(), n );
}
void EMEAWindow::toggleViewer( bool n ) {
	qDebug() << "EMEAWindow::toggleViewer(" << n << ")";
	d->toggleTool( d->viewer->chainMember(), n );
}

void EMEAWindow::addWidgetToStack( QWidget* w, bool scrollarea ) {
	qDebug() << "EMEAWindow::addWidgetToStack(" << w << ")";
	if ( scrollarea ) {
		QScrollArea* tmp = new QScrollArea( this );
		tmp->setObjectName( w->objectName() );
		tmp->setWindowIcon( w->windowIcon() );
		tmp->setWidget( w );
		tmp->setWidgetResizable( true );
		d->stack->addWidget( tmp );
	} else
		d->stack->addWidget( w );
}

void EMEAWindow::addToChainStart( NSCore::DataSetProducer* item ) {
	qDebug() << "EMEAWindow::addToChainStart(" << item << ")";
	d->producers.append( item );
	d->treemodel->addToRootItem( item );
	addLogging( item->objectName(), item, SIGNAL( countsPerSecond( double ) ) );
}

void EMEAWindow::addToChain( QList<NSCore::DataSetChainMember*> items ) {
	qDebug() << "EMEAWindow::addToChain(" << items << ")";
	d->chain.append( items );
	foreach( NSCore::DataSetChainMember* item, items ) {
		d->treemodel->addToRootItem( item );
		addLogging( item->objectName(), item, SIGNAL( countsPerSecond( double ) ) );
	}
}
void EMEAWindow::addToChain( NSCore::DataSetChainMember* item ) {
	qDebug() << "EMEAWindow::addToChain(" << item << ")";
	addToChain( QList<NSCore::DataSetChainMember*>() << item );
}
void EMEAWindow::addLogging( QString name, QObject* obj, const char* signal ) {
		d->logger->addLogger( name, obj, signal );
		//d->guilogger->addLogger( name, obj, signal );
		d->guimeter->addLogger( name, obj, signal );
}

void EMEAWindow::loadXmlFromFile() {
	//qDebug() << "EMEAWindow::loadXmlFromFile()";
	QString path = QFileDialog::getOpenFileName( this, "File to load from", QString(), "xml-Files (*.xml)" );
	if ( ! path.isEmpty() )
		loadXmlFromFile( path );
}
void EMEAWindow::loadXmlFromFile( QString path ) {
	//qDebug() << "EMEAWindow::loadXmlFromFile(" << qPrintable( path ) << ")";
	QFile f( path );
	if ( f.open( QIODevice::ReadOnly | QIODevice::Text ) ) {
		QDomDocument doc( "eposxml" );
		doc.setContent( &f );
		f.close();
		loadXml( doc );
	}
}
void EMEAWindow::loadXml( QDomDocument doc ) {
	qDebug() << "EMEAWindow::loadXml( doc )";
	QList<NSCore::QTreeItem*> items = d->getAllRootItems();
	QDomElement elem = doc.documentElement().firstChildElement();
	while ( ! elem.isNull() ) {
		foreach( NSCore::QTreeItem* item, items ) {
			qDebug() << QString( " trying %1 on %2" ).arg( elem.nodeName() ).arg( qptrdiff( item ) );
			item->fromXml( elem );
		}
		elem = elem.nextSiblingElement();
	}
}
void EMEAWindow::saveXmlToFile() {
	//qDebug() << "EMEAWindow::saveXmlToFile()";
	QString path = QFileDialog::getSaveFileName( this, "Save to file", QString(), "xml-Files (*.xml)" );
	if ( ! path.isEmpty() ) {
		QFile f( path );
		if ( f.open( QIODevice::WriteOnly | QIODevice::Text ) ) {
			QDomDocument doc( "eposxml" );
			QDomNode node = doc.createElement( "emea" );
			doc.appendChild( node );
			saveXml( node, doc );
			f.write( doc.toByteArray() );
			f.close();
		}
	}
}
void EMEAWindow::saveXml( QDomNode parentnode, QDomDocument doc ) {
	QList<NSCore::QTreeItem*> items = d->getAllRootItems();
	foreach( NSCore::QTreeItem* item, items )
		parentnode.appendChild( item->toXml( doc ) );
}

SpectrumGuiCreator* EMEAWindow::spectrumGuiCreator() { return d->spectrumguicreator; }

QStringList EMEAWindow::arguments() { return d->arguments; }

void EMEAWindow::delayedAutoStart() {
	qDebug() << "EMEAWindow::delayedAutoStart() Trying to autostart the evaluation";
	autoStart();
}

void EMEAWindow::addThread( QThread* thread ) {
	d->threads.append( thread );
}

