/*
    Copyright 2005-2008 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef POINTER_TEMPLATE_H
#define POINTER_TEMPLATE_H

#include <list>
#include <QtCore/QDebug>

namespace NSCore {

	/**
	 * @brief simple base class for the Pointer_template to have a virtual function in it
	 */
	class PointerProtection_Base
	{
		public:
			virtual void setNull() = 0;
			virtual ~PointerProtection_Base() {}
	};

	/**
	 * @brief Protect your pointers.
	 *
	 * Subclass this class if  you want your Pointer_template<Bla> to be
	 * protected against deletion.
	 *
	 * That means Bla has to be a subclass of PointerProtection_Skel and every
	 * Pointer_template<Bla> becomes zero if Bla is deleted. So you don't get
	 * dangling pointers...
	 */
	class PointerProtection_Skel
	{
		public:
			/// destructor
			/**
			 * Tells all the registered pointers that this object is deleted.
			 */
			virtual ~PointerProtection_Skel() {
				clearPointers();
			}
			/// register a pointer
			/**
			 * This one is called by Pointer_template to register itself as a
			 * pointer and get notified if this object is deleted.
			 */
			void registerPointer( PointerProtection_Base* n ) {
				_pointers.push_back( n );
			}

			/// unregister a pointer
			/**
			 * Called by Pointer_template to unregister itself from the
			 * registered pointers.
			 */
			void unregisterPointer( PointerProtection_Base* n ) {
				_pointers.remove( n );
			}

			/// Removes all pointers from the list.
			void clearPointers() {
				while ( !_pointers.empty() ) {
					_pointers.front()->setNull();
					_pointers.pop_front();
				}
			}

		protected:
			std::list<PointerProtection_Base*> _pointers;
	};

	/**
	 * @brief Template class for pointer-arithmetic.
	 *
	 * If you subclass your classes from PointerProtection_Skel you will even
	 * get protection against dangling pointers after deletion.
	 */
	template<class T> class Pointer_template : public PointerProtection_Base
	{
			/**
			 * No creation of the Pointer without the real pointer
			 */
			Pointer_template() : _p( 0 ) {}

		public:
			/**
			 * Creates a Pointer from the given pointer
			 */
			Pointer_template( T* p ) : _p( p ), _shownwarning( false ) {
				PointerProtection_Skel* tmp = dynamic_cast<PointerProtection_Skel*>( p );
				if ( tmp != 0 )
					tmp->registerPointer( this );
			}

			/**
			 * Copy constructor
			 */
			Pointer_template( const Pointer_template<T>& n ) : _p( n._p ), _shownwarning( false ) {
				//std::cerr << "\n Copyconstructor of Pointer_template isn't implemented yet!\n\n";
				PointerProtection_Skel* tmp = dynamic_cast<PointerProtection_Skel*>( _p );
				if ( tmp != 0 )
					tmp->registerPointer( this );
			}

			/// destructor
			virtual ~Pointer_template() {
				PointerProtection_Skel* tmp = dynamic_cast<PointerProtection_Skel*>( _p );
				if ( tmp != 0 )
					tmp->unregisterPointer( this );
			}

			/**
			 * operator =
			 */
			Pointer_template operator=( Pointer_template pointer ) {
				_p = pointer._p;
				PointerProtection_Skel* tmp = dynamic_cast<PointerProtection_Skel*>( _p );
				if ( tmp != 0 )
					tmp->registerPointer( this );
				return *( this );
			}

			/// Access
			T* operator* () const {
				/*if ( _p == 0 )
					qDebug() << "Pointer_template::operator*() returning 0x0!\n";
				return _p;*/
				return object();
			}
			/// Access
			T* operator->() const {
				/*if ( _p == 0 )
					qDebug() << "Pointer_template::operator->() returning 0x0!\n";
				return _p;*/
				return object();
			}
			/// Access
			T* object() const {
				if ( _p == 0 && ! _shownwarning ) {
					qWarning() << "Pointer_template::object() returning 0x0!";
					_shownwarning = true;
				}
				return _p;
			}

			virtual void setNull() {
				_p = 0;
			}

			bool isNull() const {
				if ( _p == 0 )
					return true;
				return false;
			}

		protected:
			T* _p;
		private:
			mutable bool _shownwarning;
	};

	/**
	 * @brief Comparison for pointer templates
	 *
	 * @related Pointer_template
	 */
	template< class T >
	bool operator==( Pointer_template<T> one, Pointer_template<T> two )
	{
		return one.object() == two.object();
	}

	/**
	 * @brief Comparison for pointer templates
	 *
	 * @related Pointer_template
	 */
	template< class T >
	bool operator<( Pointer_template<T> one, Pointer_template<T> two )
	{
		return one.object() < two.object();
	}

}; //NSCore

#endif // POINTER_TEMPLATE_H

