/*
    Copyright 2005-2008 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef THREAD_TEMPLATE_H
#define THREAD_TEMPLATE_H

#include <QtCore/QDebug>
#include <QtCore/QMutex>
#include <QtCore/QThread>
#include <QtCore/QCoreApplication>

namespace NSCore {

	/**
	 * @brief Run an object in its own thread
	 *
	 * This runs whatever object you have in its own thread. Its basicly the
	 * same as inheriting from QThread but easier to apply to objects since
	 * there is no need for a new class. So you can use your classes in threaded
	 * and non-threaded environments without introducing new classes or changing
	 * existing ones.
	 *
	 * The object of the thread is created at the start of run(), then a
	 * QEventLoop is running and when the thread gets stopped the object is
	 * deleted.
	 *
	 * Also if you delete the thread-object itself the eventloop is stopped and
	 * the threaded object deleted.
	 */
	template <class ThreadClass> class Thread_Template : public QThread
	{
		public:
			/**
			 * @brief Create a new thread object
			 */
			Thread_Template( QObject* p )
				: QThread( p )
				, _item( 0 )
			{
			}
			/**
			 * @brief Stop the thread and delete the thread object
			 */
			~Thread_Template() {
				quit();
				while ( isRunning() )
					usleep( 10 );
				delete _item;
			}

			/**
			 * @brief Access the threaded object
			 */
			ThreadClass* item() {
				return _item;
			}

			/**
			 * @brief wait till the thread is started and the object is created.
			 */
			bool waitForStarted() {
				//qDebug() << "Thread_Template::waitForStarted()";
				if ( isRunning() ) {
					while ( _item == 0 ) {
						//qDebug() << "Thread_Template::waitForStarted() _item =" << _item;
						QCoreApplication::processEvents();
					}
					return true;
				}
				return false;
			}

		private:
			void run() {
				_item = new ThreadClass();

				exec();

				delete _item;
				_item = 0;
			}

			ThreadClass* _item;
	};

};

#endif // THREAD_TEMPLATE_H

