/*
    Copyright 2005-2008 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#include "abstractdiskio.h"
#include "abstractdiskio.moc"

#include <QtCore/QDebug>
#include <QtCore/QFile>
#include <QtCore/QCoreApplication>
#include <QtCore/QThread>
#include <QtCore/QTime>
#include <QtCore/QTimer>

#include <actiontreeitem.h>

namespace NSDiskIO {

	AbstractReader::AbstractReader( QObject* p ) 
		: NSCore::DataSetProducer( p )
		, config( new NSCore::NamedParameterTreeItem( "Parameters", QVariant::Invalid, this ) )
		, _datasetsToRead( 0 )
		, _datasetsWritten( 0 )
		, _timer( new QTimer( this ) )
	{
		qDebug() << "AbstractReader::AbstractReader(" << p << ")";
		config->addParameter( "File to read from", QVariant::Url );
		config[ "File to read from" ]->setGuiHint( NSCore::NamedParameterTreeItem::UrlRead );
		connect( config[ "File to read from" ], SIGNAL( valueChanged( QString ) ), this, SLOT( setFile( QString ) ) );
		NSCore::ActionTreeItem* action = new NSCore::ActionTreeItem( "Read One Dataset", this, ":/icons/kde/cr22-action-player_fwd.png", this, SLOT( readDataSet() ) );
		new NSCore::ActionTreeItem( "Read Ten Datasets", action, this, SLOT( readTenDataSets() ) );
		new NSCore::ActionTreeItem( "Read 100 Datasets", action, this, SLOT( readHundredDataSets() ) );
		new NSCore::ActionTreeItem( "Read All Datasets", this, ":/icons/kde/cr22-action-player_end.png", this, SLOT( readAllDataSets() ) );
		new NSCore::ActionTreeItem( "Stop Reading", this, ":/icons/kde/cr22-action-player_stop.png", this, SLOT( stopReading() ) );
		new NSCore::ActionTreeItem( "Reset file to beginning", this, ":/icons/kde/cr22-action-player_start.png", this, SLOT( reset() ) );

		NSCore::NamedParameterTreeItem* counter = new NSCore::NamedParameterTreeItem( "Datasets read from file", QVariant::Int, this );
		counter->setGuiHint( NSCore::NamedParameterTreeItem::ReadAllowed );
		connect( this, SIGNAL( readDataSets( int ) ), counter, SLOT( setValue( int ) ) );

		_timer->setInterval( 0 );
		_timer->setSingleShot( true );
		connect( _timer, SIGNAL( timeout() ), this, SLOT( privateReadPendingDataSets() ) );

		QTimer::singleShot( 1, this, SLOT( privateLateInit() ) );
	}
	AbstractReader::~AbstractReader() {
	}
	void AbstractReader::privateLateInit() {
		config[ "File to read from" ]->setUrlFilter( filter() );
	}

	void AbstractReader::readDataSet() {
		readNDataSets( 1 );
	}
	void AbstractReader::readAllDataSets() {
		//qDebug() << "AbstractReader::readAllDataSets()";
		//qDebug() << " _datasetsToRead =" << _datasetsToRead << " timer.isRunning() =" << _timer->isActive();
		if ( _datasetsToRead == 0 ) {
			//qDebug() << " Starting to read all datasets.";
			_datasetsToRead = -1;
			_timer->start();
		}
		//qDebug() << " _datasetsToRead =" << _datasetsToRead << " timer.isRunning() =" << _timer->isActive();
	}

	void AbstractReader::readNDataSets( int n ) {
		if ( _datasetsToRead == -1 )
			return;
		if ( _datasetsToRead == 0 )
			_timer->start();
		_datasetsToRead += n;
	}
	void AbstractReader::privateReadPendingDataSets() {
		if ( _datasetsToRead == 0 )
			return;
		if ( atEnd() )
			return;
		//if ( thread() != QThread::currentThread() )
		//	qWarning() << "AbstractReader::privateReadPendingDataSets() Not running in my thread!!!";
		readSingleDataSet();
		emit readDataSets( ++_datasetsWritten );
		if ( _datasetsToRead > 0 )
			_datasetsToRead--;
		_timer->start();
	}
	void AbstractReader::readTenDataSets() {
		readNDataSets( 10 );
	}
	void AbstractReader::readHundredDataSets() {
		readNDataSets( 100 );
	}
	QDomNode AbstractReader::saveToXML( QDomDocument doc ) {
		return doc.createElement( "reader" );
	}
	bool AbstractReader::restoreFromXML( QDomNode node ) {
		if ( node.nodeName() == "reader" ) {
			QDomElement elem = node.toElement();
			if ( elem.hasAttribute( "filename" ) )
				setFile( elem.attribute( "filename" ) );
			return true;
		}
		return false;
	}
	QVariant AbstractReader::guiIdentifier( int role ) const {
		switch ( role ) {
			case Qt::DecorationRole:
				return ":/icons/cr48-action-load_pulses.png";
				break;
			default:
				return NSCore::DataSetProducer::guiIdentifier( role );
		}
	}

	AbstractWriter::AbstractWriter( QObject* p )
		: NSCore::DataSetChainMember( p )
		, config( new NSCore::NamedParameterTreeItem( "Parameters", QVariant::Invalid, this ) )
		, _number_of_written_datasets( 0 )
	{
		setObjectName( "AbstractDiskWriter" );

		NSCore::NamedParameterTreeItem* written = new NSCore::NamedParameterTreeItem( "Written DataSets", 0, this );
		written->setGuiHint( NSCore::NamedParameterTreeItem::Counter );
		connect( this, SIGNAL( wroteDataSet( int ) ), written, SLOT( setValue( int ) ) );

		config->addParameter( "Filename to write to", QVariant::Url );
		config[ "Filename to write to" ]->setGuiHint( NSCore::NamedParameterTreeItem::UrlWrite );
		connect( config[ "Filename to write to" ], SIGNAL( valueChanged( QString ) ), this, SLOT( setFile( QString ) ) );

		QTimer::singleShot( 1, this, SLOT( privateLateInit() ) );
	}
	bool AbstractWriter::dataArrived( NSCore::AcqisitionDataSet& data ) {
		//qDebug() << "AbstractWriter::dataArrived(" << &data << ")";
		writeDataSet( data );
		++_number_of_written_datasets;
		emit wroteDataSet( _number_of_written_datasets );
		return true;
	}
	void AbstractWriter::privateLateInit() {
		config[ "Filename to write to" ]->setUrlFilter( filter() );
	}

	void AbstractWriter::resetCounter() {
		_number_of_written_datasets = 0;
	}


}

