/*
    Copyright 2008 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>
                        Nicki Hinsche <nicvoc@freenet.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#include "dopplerproducer.h"
#include "dopplerproducer.moc"

#include <cmath>

#include <QtCore/QDateTime>
#include <QtCore/QTimer>
#include <QtCore/QCoreApplication>
#include <QtCore/QThread>
#include <QtCore/QFile>
#include <QtCore/QDataStream>

#include <actiontreeitem.h>
#include <namedparametertreeitem.h>

using namespace NSCore;

class DopplerProducer::DopplerProducer_Private {
	public:

		bool ownthread;

		int counter, toread, numbers;
		QString parentpath, prefix, date;

		QString getFilename( int ch, int set ) {
			QString channel = QString::number( ch ).rightJustified( 2, '0' );
			for ( int folder=1; folder<100; ++folder ) {
				QString name = QString( "%1/%2-%3_CH%4/Folder.%5/%2_CH%4-%6.sig" ).arg( parentpath ).arg( prefix ).arg( date ).arg( channel ).arg( QString::number( folder ).rightJustified( 5, '0' ) ).arg( QString::number( set ).rightJustified( numbers, '0' ) );
				if ( QFile::exists( name ) )
					return name;
			}
			return "";
		}

		NSCore::NamedParameterTreeItem* path;
		NSCore::NamedParameterTreeItem* togenerate;

		DopplerProducer_Private() : counter( 0 ), toread( 0 ) {
		}
};

DopplerProducer::DopplerProducer( QObject* p )
	: DataSetProducer( p )
	, d( new DopplerProducer_Private )
{
	//qDebug() << endl << "DopplerProducer::DopplerProducer(" << p << ")";
	setObjectName( "DopplerProducer" );

	d->path = new NamedParameterTreeItem( "Path", QVariant::String, this );
	connect( d->path, SIGNAL( valueChanged( const QString& ) ), this, SLOT( pathChanged( const QString& ) ) );

	new NSCore::ActionTreeItem( "Generate OneDataSet", this, this, SLOT( readSingleDataSet() ) );
	NSCore::ActionTreeItem* multiple = new NSCore::ActionTreeItem( "Generate N Datasets", this, this, SLOT( readNDataSets() ) );
	d->togenerate = new NSCore::NamedParameterTreeItem( "to read", 1, multiple );

	srand( QDateTime::currentDateTime().toTime_t() );
	d->ownthread = ( thread() != QCoreApplication::instance()->thread() );

	//qDebug() << "DopplerProducer::DopplerProducer() finished" << endl;
}
DopplerProducer::~DopplerProducer() {
}

void DopplerProducer::readNDataSets() {
	readNDataSets( d->togenerate->toInt() );
}

void DopplerProducer::pathChanged( const QString& _path ) {
	QString path = _path;
	qDebug() << "DopplerProducer::pathChanged(" << path << ")";
	if ( path.endsWith( "/" ) )
		path.chop( 1 );
	d->parentpath = path.section( "/", 0, -2 );
	QString name = path.section( "/", -1, -1 );
	d->prefix = name.section( "-", 0, -6 );
	d->date = name.section( "-", -5, -1 ).section( "_", 0, 1 );
	//qDebug() << d->prefix << " " << d->date;

	d->numbers = 1;
	while ( ! QFile::exists( d->getFilename( 1, 1 ) ) && d->numbers < 10 ) {
		//qDebug() << "DopplerProducer::pathChanged()  numbers is" << d->numbers << " file(" << d->getFilename( 1,1 ) << ") seems to not exist";
		d->numbers++;
	}

	d->counter = 1;
	qDebug() << d->getFilename( 1, 1 );
}

void DopplerProducer::readNDataSets( int n ) {
	if ( d->toread == 0 )
		QTimer::singleShot( 100, this, SLOT( readDataSet() ) );
	d->toread += n;
}
void DopplerProducer::readSingleDataSet() {
	readNDataSets( 1 );
}

bool DopplerProducer::restoreFromXML( QDomNode node ) {
	if ( node.nodeName() == "reader" && node.isElement() && node.toElement().attribute( "type" ) == "dopplerproducer" )
		return true;
	return false;
}

///the real work

void DopplerProducer::readDataSet() {

	if ( QThread::currentThread() != thread() ) {
		QTimer::singleShot( 10, this, SLOT( readDataSet() ) );
		return;
	}

	//qDebug() << "DopplerProducer::readDataSet()";

	const int channels = 2;

	QMap< int, QVector<double> > data;
	int skip=512;
	qint16 buffer=0;

	for ( int i=0; i<channels; ++i ) {

		//qDebug() << " opening " << d->getFilename( i+1, d->counter );
		QFile file( d->getFilename( i+1, d->counter ) );
		if ( ! file.open( QIODevice::ReadOnly ) ) {
			qWarning() << "  Could not open " << d->getFilename( i+1, d->counter );
			return;
		}
		QDataStream _stream( &file );
		_stream.skipRawData(skip);
		_stream.setByteOrder( QDataStream::LittleEndian );

		while ( ! _stream.atEnd() ) {
			_stream >> buffer;
			data[ i ].push_back( buffer );
		}
	}

	AcqisitionDataSet dataset( channels, qMax( data[ 0 ].size(), data[ 1 ].size() ) );

	const double imax = pow( 2,14 );//just normalize
	for ( int i=0; i<dataset.channels(); ++i ) {
		double* ch = dataset.getRawChannel( i );

		for ( int j=0; j<dataset.samples(); ++j ) {
			ch[ j ] = data[ i ][ j ] / imax;
		}
		dataset.updateMinMax();

		double range = dataset.result( QString( "channel %1/max" ).arg( i ) )->toDouble() - qMin( dataset.result( QString( "channel %1/min" ).arg( i ) )->toDouble(), 0.0 );
		dataset.addResult( QString( "channel %1/fullscale" ).arg( i ), range*2 );
	}

	enqueue( dataset );

	d->counter++;
	d->toread--;
	if ( d->toread > 0 ) {
		if ( d->ownthread )
			QTimer::singleShot( 0, this, SLOT( readDataSet() ) );
		else
			QTimer::singleShot( 10, this, SLOT( readDataSet() ) );
	}
}


