/*
    Copyright 2005-2008 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#include "osciwidget_v3.h"
#include "osciwidget_v3.moc"

#include "osciwidget_v3_private.h"
#include "osciwidget_v3_private.moc"

#include <math.h>
#include <QtCore/QMutex>
#include <QtCore/QMutexLocker>
#include <QtGui/QPainter>
#include <QtGui/QVBoxLayout>
#include <QtGui/QInputDialog>
#include <QtGui/QAction>
#include <QtGui/QMenu>
#include <QtGui/QContextMenuEvent>
#include <QtCore/QTimer>

OsciWidgetV3_ChainMember::OsciWidgetV3_ChainMember( QObject* p ) : NSCore::DataSetChainMember( p ) {
	_slowdown.start();
}
bool OsciWidgetV3_ChainMember::dataArrived( NSCore::AcqisitionDataSet& d ) {
	//qDebug() << "OsciWidgetV3_ChainMember::dataArrived(" << d << ")";
	if ( _slowdown.elapsed() > 200 ) {
		emit data( d );
		_slowdown.restart();
	}
	return true;
}

OsciWidgetV3::OsciWidgetV3( QWidget* p )
	: QWidget( p )
	, _layout( new QVBoxLayout( this ) )
	, _options( new OsciWidgetV3_Options( this ) )
	, _chainmember( new OsciWidgetV3_ChainMember( this ) )
{
	connect( _chainmember, SIGNAL( data( NSCore::AcqisitionDataSet& ) ), this, SLOT( dataAvailable( NSCore::AcqisitionDataSet& ) ) );

	setMinimumSize( 100,100 );
	_layout->addWidget( _options, 1 );
	connect( _options->spinboxChannels, SIGNAL( valueChanged( int ) ), this, SLOT( setChannels( int ) ) );
	setChannels( _options->spinboxChannels->value() );

	setObjectName( "Oscilloscope" );
	setWindowIcon( QIcon( ":/icons/cr48-action-osci.png" ) );
}
OsciWidgetV3::~OsciWidgetV3() {
}

void OsciWidgetV3::dataAvailable( NSCore::AcqisitionDataSet& data ) {
	//qDebug() << "OsciWidgetV3::dataAvailable( data )";
	for ( int i=0; i<_oscis.size(); ++i ) {
		_oscis[ i ]->data( data );
	}
}

void OsciWidgetV3::setChannels( int n ) {
	while ( _oscis.size() > n ) {
		delete _oscis.takeLast();
	}
	while ( _oscis.size() < n ) {
		_oscis.push_back( new OsciWidgetV3_Osci( this ) );
		_layout->addWidget( _oscis.last(), 255 );
		_oscis.last()->setChannel( _oscis.size()-1 );
	}
}

NSCore::DataSetChainMember* OsciWidgetV3::chainMember() { return _chainmember; }



OsciWidgetV3_Osci::OsciWidgetV3_Osci( QWidget* p )
	: QWidget( p )
	, fullscale( 0 )
	, offset( 0 )
	, _channel( 0 )
	, _generation( 0 )
	, _mutex( new QMutex() )
{
	QAction* channelselect = new QAction( "Select Channel...", this );
	connect( channelselect, SIGNAL( triggered() ), this, SLOT( enterChannel() ) );
	addAction( channelselect );
	QAction* generationselect = new QAction( "Select Generation...", this );
	connect( generationselect, SIGNAL( triggered() ), this, SLOT( enterGeneration() ) );
	addAction( generationselect );

	QMenu* x_actions = new QMenu( "Horizontal Data", this );
	addAction( x_actions->menuAction() );
	foreach ( QString tmp, QStringList() << "interpolated threshold at" << "interpolated minimum at" << "interpolated fraction point at" ) {
		QAction* action = new QAction( tmp, x_actions );
		action->setCheckable( true );
		action->setChecked( true );
		x_actions->addAction( action );
		_actions_x[ tmp ] = action;
	}
	QMenu* y_actions = new QMenu( "Horizontal Data", this );
	addAction( y_actions->menuAction() );
	foreach ( QString tmp, QStringList() << "interpolated minimum" << "interpolated fraction point" << "zeroline" ) {
		QAction* action = new QAction( tmp, y_actions );
		action->setCheckable( true );
		action->setChecked( true );
		y_actions->addAction( action );
		_actions_y[ tmp ] = action;
	}
}
OsciWidgetV3_Osci::~OsciWidgetV3_Osci() {
	delete _mutex;
}

void OsciWidgetV3_Osci::data( NSCore::AcqisitionDataSet& data ) {
	if ( ! data.channels() > _channel )
		_channel = data.channels() - 1;

	QMutexLocker locker( _mutex );

	int generation = _generation;
	QString prefix = data.getPrefix( generation );
	//if ( data.generations() <= generation )
	//	generation = channel->currentGeneration();

	if ( data.result( prefix + QString( "/channel %1/fullscale" ).arg( _channel ) )->isValid() ) {
		fullscale = data.result( prefix + QString( "/channel %1/fullscale" ).arg( _channel ) )->toDouble();
		offset    = data.result( prefix + QString( "/channel %1/offset" ).arg( _channel ) )->toDouble();
	} else {
		double a = data.result( prefix + QString( "/channel %1/max" ).arg( _channel ) )->toDouble();
		double b = data.result( prefix + QString( "/channel %1/min" ).arg( _channel ) )->toDouble();
		//qDebug() << "max =" << a << " min =" << b;
		fullscale = a - b;
		offset = fullscale / -2.0 - b;
	}
	//qDebug() << "Fullscale =" << fullscale << " With Offset =" << offset;

	_data.clear();
	int inc = 1;
	if ( data.samples( generation ) > width() )
		inc = data.samples( generation ) / width();

	double* in = data.getRawChannel( _channel, generation );
	for ( int i=0; i<data.samples( generation ); i+=inc )
		_data.push_back( in[ i ] );

	foreach( QString name, _actions_x.keys() ) {
		if ( _actions_x[ name ]->isChecked() ) {
			QVariant* tmp = data.result( QString( "%1/channel %2/%3" ).arg( prefix ).arg( _channel ).arg( name ) );
			if ( tmp->isValid() )
				_addons_x[ name ] = tmp->toDouble();
		} else
			_addons_x.remove( name );
	}
	foreach( QString name, _actions_y.keys() ) {
		if ( _actions_y[ name ]->isChecked() ) {
			QVariant* tmp = data.result( QString( "%1/channel %2/%3" ).arg( prefix ).arg( _channel ).arg( name ) );
			if ( tmp->isValid() )
				_addons_y[ name ] = tmp->toDouble();
		} else
			_addons_y.remove( name );
	}

	update();
}

void OsciWidgetV3_Osci::paintEvent( QPaintEvent* ) {
	QMutexLocker locker( _mutex );
	QPainter p( this );

	p.save();

	double dx = width() / double( _data.size() );
	double dy = - height() / fullscale;

	p.translate( 0, height()/2 + dy*offset );

	/*p.setPen( QColor( 150,150,150 ) );
	p.drawText( QPointF( 0, 10 ),
			QString( "Fullscale: %1, Offset: %2, results in [%3,%4], dy: %5" )
			.arg( fullscale )
			.arg( offset )
			.arg( ( -offset - fullscale/2 ) * dy )
			.arg( ( -offset + fullscale/2 ) * dy )
			.arg( dy )
			);*/

	p.setPen( QColor( 0,0,0 ) );

	p.drawLine( 0, 0, width(), 0 );

	if ( _data.size() > 0 ) {
		QPointF oldpoint( 0, _data[ 0 ] * dy );
		QPointF newpoint;
		for ( int i=1; i<_data.size(); ++i ) {
			newpoint = QPointF(  i*dx, _data[ i ] * dy );
			p.drawLine( oldpoint, newpoint );
			oldpoint = newpoint;
		}
	}

	p.setPen( QColor( 100,100,255 ) );
	{
		QMapIterator<QString,double> it( _addons_x );
		int i=0;
		while ( it.hasNext() ) {
			it.next();
			if ( it.value() != 0 ) {
				int x = int( dx * it.value() );
				p.drawLine( x, int( -fullscale*dy ), x, int( fullscale*dy ) );
				p.drawText( x+10, 30 - i*20, QString( "%1: %2" ).arg( it.key() ).arg( it.value() ) );
				++i;
			}
		}
	}
	{
		QMapIterator<QString,double> it( _addons_y );
		while ( it.hasNext() ) {
			it.next();
			if ( it.value() != 0 ) {
				int y = int( dy * it.value() );
				p.drawLine( 0, y, width(), y );
				p.drawText( 10, y-10, QString( "%1: %2" ).arg( it.key() ).arg( it.value() ) );
			}
		}
	}

	p.restore();
	p.drawRect( rect().adjusted( 0,0, -1, -1 ) );
}

void OsciWidgetV3_Osci::setChannel( int n ) { _channel = n; }

void OsciWidgetV3_Osci::contextMenuEvent( QContextMenuEvent* ev ) {
	QMenu::exec( actions(), ev->globalPos() );
}

void OsciWidgetV3_Osci::enterChannel() {
	_channel = QInputDialog::getInteger( this,
			"Select the new channel to show",
			"Channel to show:",
			_channel, 0 );
}
void OsciWidgetV3_Osci::enterGeneration() {
	_generation = QInputDialog::getInteger( this,
			"Select the new generation to show",
			"Generation to show:",
			_generation, 0 );
}


OsciWidgetV3_Options::OsciWidgetV3_Options( QWidget* p )
	: QFrame( p )
{
	setupUi( this );
}

