/*
    Copyright 2005-2008 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#include "filters.h"
#include "analyzers.h"

//#include <math.h>
#include <QtCore/QDebug>
#include <filterexpression.h>

PeakFilter::PeakFilter( NSCore::QTreeItem* p ) : NSEvaluation::Filter( p ) {
	_config->addParameter( "threshold", -0.1 );
	_config[ "threshold" ]->setMin( -50.0 );
	_config[ "threshold" ]->setMax( 50.0 );
	_config->addParameter( "peaks", 1 );
	_config[ "peaks" ]->setMin( 0 );
	_config[ "peaks" ]->setMax( 20 );
}

bool PeakFilter::filter( NSCore::AcqisitionDataSet& data ) {
	//qDebug() << "PeakFilter::filter( data )";
	double threshold = _config->getParameter( "threshold" )->toDouble();
	int peakcount = _config->getParameter( "peaks" )->toInt();
	int _updown, _downup;
	bool _up = false;
	bool _result = true;
	for ( int c=0; c<data.channels(); ++c ) {

		double* ch = data.getRawChannel( c );

		_updown = _downup = 0;

		if ( ch[ 0 ] > threshold )
			_up = true;

		for ( int i=1; i<data.samples(); ++i ) {
			double sample = ch[ i ];
			if ( _up && sample < threshold ) {
				++ _updown;
				_up = false;
			} else if ( !_up && sample > threshold ) {
				++ _downup;
				_up = true;
			}
		}

		if ( _updown == _downup && _downup > 0 && _result && _downup <= peakcount )
			;
		else
			_result = false;

	}

	//qDebug() << " returning" << _result;
	if ( _result )
		data.addResult( "peak coincidence", _result );
	return _result;
}

TwoChannelDisc::TwoChannelDisc( NSCore::QTreeItem* p )
	: NSEvaluation::Filter( p )
	, _analyzer( new InterpolatedMinimumAnalyzer( _config.object() ) )
{
	_config->addParameter( "Channel 0/lower bound", -5.0 );
	_config[ "Channel 0/lower bound" ]->setMin( -50.0 );
	_config[ "Channel 0/lower bound" ]->setMax( 50.0 );
	_config->addParameter( "Channel 0/upper bound", 5.0 );
	_config[ "Channel 0/upper bound" ]->setMin( -50.0 );
	_config[ "Channel 0/upper bound" ]->setMax( 50.0 );
	_config->addParameter( "Channel 1/lower bound", -5.0 );
	_config[ "Channel 1/lower bound" ]->setMin( -50.0 );
	_config[ "Channel 1/lower bound" ]->setMax( 50.0 );
	_config->addParameter( "Channel 1/upper bound", 5.0 );
	_config[ "Channel 1/upper bound" ]->setMin( -50.0 );
	_config[ "Channel 1/upper bound" ]->setMax( 50.0 );
	_config[ "Channel 0" ]->setToolTip( "Minimum and Maximum of the window for channel 0." );
	_config[ "Channel 1" ]->setToolTip( "Minimum and Maximum of the window for channel 1." );
}

bool TwoChannelDisc::filter( NSCore::AcqisitionDataSet& data ) {
	bool ch[ data.channels() ];
	for ( int i=0; i<data.channels() && i<2; ++i ) {
		ch[ i ] = false;
		if ( data.result( QString( "channel %1/interpolated minimum" ).arg( i ) ) == 0 )
			_analyzer->analyze( data );
		double value = data.result( QString( "channel %1/interpolated minimum" ).arg( i ) )->toDouble();
		if ( value > _config->getParameter( QString( "Channel %1/lower bound" ).arg( i ) )->toDouble() )
			if ( value < _config->getParameter( QString( "Channel %1/upper bound" ).arg( i ) )->toDouble() )
					ch[ i ] = true;
	}
	bool ret=true;
	for ( int i=0; i<data.channels(); ++i )
		ret = ret & ch[ i ];
	return ret;
}

TwoChannelXORDisc::TwoChannelXORDisc( NSCore::QTreeItem* p )
	: NSEvaluation::Filter( p )
	, _analyzer( new InterpolatedMinimumAnalyzer( _config.object() ) )
{
	_config->addParameter( "Window 0/upper bound", 5.0 );
	_config[ "Window 0/upper bound" ]->setMin( -50.0 );
	_config[ "Window 0/upper bound" ]->setMax( 50.0 );
	_config->addParameter( "Window 0/lower bound", -5.0 );
	_config[ "Window 0/lower bound" ]->setMin( -50.0 );
	_config[ "Window 0/lower bound" ]->setMax( 50.0 );
	_config->addParameter( "Window 1/upper bound", 5.0 );
	_config[ "Window 1/upper bound" ]->setMin( -50.0 );
	_config[ "Window 1/upper bound" ]->setMax( 50.0 );
	_config->addParameter( "Window 1/lower bound", -5.0 );
	_config[ "Window 1/lower bound" ]->setMin( -50.0 );
	_config[ "Window 1/lower bound" ]->setMax( 50.0 );
	_config[ "Window 0" ]->setToolTip( "Minimum and Maximum for the first window" );
	_config[ "Window 1" ]->setToolTip( "Minimum and Maximum for the second window" );
}

bool TwoChannelXORDisc::filter( NSCore::AcqisitionDataSet& data ) {
	bool windows_reached[ data.channels() ];
	double mins[ data.channels() ], maxs[ data.channels() ];
	for ( int i=0; i<data.channels() && i<2; ++i ) {
		windows_reached[ i ] = false;
		mins[ i ] = _config->getParameter( QString( "Window %1/lower bound" ).arg( i ) )->toDouble();
		maxs[ i ] = _config->getParameter( QString( "Window %1/upper bound" ).arg( i ) )->toDouble();
	}

	for ( int i=0; i<data.channels(); ++i ) {

		if ( data.result( QString( "channel %1/interpolated minimum" ).arg( i ) ) == 0 )
			_analyzer->analyze( data );
		double value = data.result( QString( "channel %1/interpolated minimum" ).arg( i ) )->toDouble();

		for ( int j=0; j<data.channels(); ++j ) {
			if ( windows_reached[ j ] == false && ( value > mins[ j ] ) && ( value < maxs[ j ] ) )
				windows_reached[ j ] = true;
		}

	}

	bool ret = true;
	for ( int i=0; i<data.channels(); ++i )
		ret = ret & windows_reached[ i ];
	return ret;
}

SimpleDisc::SimpleDisc( NSCore::QTreeItem* p )
	: NSEvaluation::Filter( p )
	, _analyzer( new InterpolatedMinimumAnalyzer( _config.object() ) )
{
	_config->addParameter( "lower bound", -5.0 );
	_config[ "lower bound" ]->setMin( -50.0 );
	_config[ "lower bound" ]->setMax( 50.0 );
	_config->addParameter( "upper bound", 5.0 );
	_config[ "upper bound" ]->setMin( -50.0 );
	_config[ "upper bound" ]->setMax( 50.0 );
}

bool SimpleDisc::filter( NSCore::AcqisitionDataSet& data ) {
	double lower = _config[ "lower bound" ]->toDouble();
	double upper = _config[ "upper bound" ]->toDouble();
	bool result = true;
	for ( int i=0; i<data.channels() && result; ++i ) {
		if ( data.result( QString( "channel %1/interpolated minimum" ).arg( i ) ) == 0 )
			_analyzer->analyze( data );
		double min = data.result( QString( "channel %1/interpolated minimum" ).arg( i ) )->toDouble();
		if ( ( min > lower ) && ( min < upper ) && ( result == true ) )
			result = true;
		else
			result = false;
	}
	return result;
}

MathFilter::MathFilter( NSCore::QTreeItem* p )
	: NSEvaluation::Filter( p )
	, _filter( new NSEvaluation::FilterExpression( p ) )
{
}

bool MathFilter::filter( NSCore::AcqisitionDataSet& data ) {
	return !isnan( _filter->value( data ) );
}

