/*
    Copyright 2005-2008 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef INTERPOLATION_HELPERS_H
#define INTERPOLATION_HELPERS_H

/**
 * @brief A coordinate
 * @internal
 *
 * This simple struct ties x and y together to a coordinate.
 */
struct coord {
	coord( double _x, double _y ) : x( _x ), y( _y ) { }
	coord() : x( 0 ), y( 0 ) { }
	double x;
	double y;
};

/**
 * @brief Polynom interpolation
 *
 * This is taken from numerical recipes.
 *
 * This class does a polynomical interpolation of n-th order on the data
 * provided.
 *
 * You can set the order of interpolation either during creation or at a later
 * time. Of course it also can be changed at any time.
 */
class PolynomInterpolate
{
	public:
		/**
		 * @brief create a new polynom interpolation
		 *
		 * @param order The order of the interpolation
		 */
		PolynomInterpolate( int order ) {
			setOrder( order );
		}
		/**
		 * @brief create a new polynom interpolation
		 *
		 * The interpolation will be 1st order but that can be changed at any
		 * time. @see setOrder( int )
		 */
		PolynomInterpolate() {
			setOrder( 1 );
		}

		void setOrder( int n ) { _order = n+1; }
		void setMiddle( int n ) { _begin = n-( _order/2 ) + 1; }
		void setBegin( int n ) { _begin = n; }
		void setData( double* n ) { _ya = n; }

		/**
		 * interpolate value x
		 */
		double polint( double x );

	protected:
		double* _ya;
		int _begin;
		int _order;
};

/**
 * @brief Search for the minimum of a function
 *
 * This is also taken from Numerical Recipes.
 *
 * It searches with nested intervals for an interpolated minimum around a given
 * value. It should be given the index of the smallest known value.
 *
 * You have to implement function() to do the actual interpolation.
 */
class Minimum
{
	public:
		Minimum() : _minimum_at_int( 0 ) {
		}

		virtual ~Minimum() {}

		void setMinimum( int n ) {
			_minimum_at_int = n;
		}

		/**
		 * @brief the actual function to be minimized
		 *
		 * Implement this to evaluate the function you are searching a minimum
		 * for.
		 */
		virtual double function( double ) =0;

		/**
		 * @brief returns the coordinates of the minimum next to the one set
		 * with setMinimum
		 */
		coord golden_minimum();
		/**
		 * @brief return the minimum next to the given one
		 */
		coord golden_minimum( int min ) {
			setMinimum( min );
			return golden_minimum();
		}

	private:
		int _minimum_at_int;
};

class Maximum
{
	public:
		Maximum() : _maximum_at_int( 0 ) {
		}

		void setMaximum( int n ) { _maximum_at_int = n; }

		virtual double function( double ) =0;

		coord golden_maximum();
		coord golden_maximum( int max ) {
			setMaximum( max );
			return golden_maximum();
		}

	private:
		int _maximum_at_int;
};

/**
 * @brief Cubic-Spline-Interpolation
 *
 * This is another algorithm taken from Numerical Recipes
 */
class Spline
{
	public:
		Spline();
		Spline( double *y, int n, double yp1, double ypn );
		virtual ~Spline();

		double splint( double x );

		void setData( double* y, int n );
		void setDerivates( double yp1, double ypn ) { _yp1 = yp1; _ypn = ypn; }

		void debug() const;

	protected:
		virtual double y( int n ) =0;
		double *_y;
		int _n;

	private:
		void createTable();

		double *_y2;
		double _yp1, _ypn;
		double *_u;
};

#endif // INTERPOLATION_HELPERS_H

