/*
    Copyright 2005-2008 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef RESULTTOSPECTRUM_H
#define RESULTTOSPECTRUM_H

#include <datasetchain.h>
#include <abstracttreeitem.h>
#include <mathexpression.h>
#include <QtCore/QObject>
#include <QtXml/QDomNode>

class ResultsToSpectrum;
class AbstractSpectrumModel;

class QTimer;

/**
 * @brief A path of expressions together with a spectrummodel
 *
 * @todo docs
 */
class ResultAction : public NSCore::MathTreeRoot
{
	Q_OBJECT
	friend class ResultsToSpectrum;
	public:
		ResultAction( ResultsToSpectrum* parent =0 );
		~ResultAction();

	public slots:
		void data( NSCore::AcqisitionDataSet );

	public:

		QString guiString() const { return QString( "Spectrum: %1" ).arg( objectName() ); }

		QStringList guiActions() const { return NSCore::MathTreeRoot::guiActions() << "Remove Me"; }
		void guiExecAction( QString, NSCore::GuiServer_Interface* );

	signals:
		void stopMe();

	private:
		QDomNode saveToXML( QDomDocument );
		bool restoreFromXML( QDomNode );
		QString debug() const { return QString( "ResultAction: %1" ).arg( objectName() ); }
		AbstractSpectrumModel* _model;
};

/**
 * @brief Sends results into a spectrummodel
 *
 * This class sends results into a SpectrumModel for further analysis and
 * storage.
 *
 * It will also be expanded to store into other models as well ( network-send,
 * sql-database, etc ).
 *
 * The results to be written into the storage can be defined by an own language
 * containing result,add,neg,smb,precision. Use and explore it online by
 * accessing this objects via TreeItemModel and QTreeView...
 */
class ResultsToSpectrum : public NSCore::DataSetChainMember
{
	Q_OBJECT
	public:
		/// constructor
		ResultsToSpectrum( QObject* =0 );
		/// destructor
		~ResultsToSpectrum();

	public slots:
		/**
		 * @brief Removes all paths
		 */
		void removeAllPaths();

		/**
		 * @brief Access the created SpectrumModels by name
		 */
		AbstractSpectrumModel* spectrum( QString name );
		/**
		 * @brief returns a list of the current spectra
		 */
		QStringList spectra() const;

		/**
		 * @brief Create path and spectrum by XML
		 */
		void addPath( QString xml );

		//@{
		/**
		 * @brief name and actions for the gui
		 */
		QStringList guiActions() const { return NSCore::QTreeItem::guiActions() << "Add Spectrum"; }
		void guiExecAction( QString, NSCore::GuiServer_Interface* );
		QVariant guiIdentifier( int ) const;
		//@}

	signals:
		/**
		 * @brief A new spectrum is created.
		 */
		void newSpectrum( AbstractSpectrumModel* );

		/**
		 * @brief Send the data to the sub-threads.
		 */
		void data( NSCore::AcqisitionDataSet );

	private:

		bool dataArrived( NSCore::AcqisitionDataSet& );

		/// debug for NSCore::QTreeItem
		QString debug() const { return "ResultsToSpectrum"; }

		QDomNode saveToXML( QDomDocument doc ) { return doc.createElement( "resultstospectrum" ); }

		bool restoreFromXML( QDomNode );
		bool createFromXML( QDomNode );

};

#endif // RESULTTOSPECTRUM_H

