/*
    Copyright 2005-2008 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#include "spectrummodel.h"
#include "spectrummodel.moc"

#include <math.h>
#include <QtCore/QDebug>
#include <QtCore/QFile>
#include <QtCore/QTextStream>

#include <modeltest.h>

AbstractSpectrumModel::AbstractSpectrumModel( QObject* p )
	: QAbstractItemModel( p )
	, _config( new NSCore::NamedParameterTreeItem( "options", QVariant::Invalid, 0 ) )
	, _changed_index_timer( new QTimer( this ) )
	, _autosavetimer( new QTimer( this ) )
{
	_changed_index_timer->start( 1000 );
	connect( _changed_index_timer, SIGNAL( timeout() ), this, SLOT( changedIndexTimerTimeout() ) );

	connect( _autosavetimer, SIGNAL( timeout() ), this, SLOT( autoSave() ) );

	_config->addParameter( "Autosave", QVariant::Bool );
	_config->addParameter( "Autosave/Filename", "autosave-DATE.csv" );
	_config->getParameter( "Autosave/Filename" )->setToolTip( "Variables to be extended: DATE - current date-time combination, MTIME - start time of the spectrum since last reset or since creation of the spectrum." );
	_config->addParameter( "Autosave/Interval", 10 );
	_config->getParameter( "Autosave/Interval" )->setRange( 1, 60 );
	connect( _config->getParameter( "Autosave" ), SIGNAL( valueChanged() ), this, SLOT( autoSaveChanged() ) );
}

QModelIndex AbstractSpectrumModel::parent( const QModelIndex& ) const {
	return QModelIndex();
}
void AbstractSpectrumModel::changedIndexTimerTimeout() {
	if ( !_changed_index.isEmpty() ) {
		QMutexLocker lock( &_changed_index_lock );
		qSort( _changed_index );
		emit dataChanged( _changed_index.first(), _changed_index.last() );
		_changed_index.clear();
	}
}

void AbstractSpectrumModel::autoSaveChanged() {
	//qDebug() << "AbstractSpectrumModel::autoSaveChanged()";
	//_config->getParameter( "Autosave" )->debugPrint();
	static bool lastactive = false;
	static int lastinterval = 0;
	if ( lastinterval != _config->getParameter( "Autosave/Interval" )->toInt() ) {
		lastinterval = _config->getParameter( "Autosave/Interval" )->toInt();
		_autosavetimer->stop();
		lastactive = false;
	}
	if ( lastactive != _config->getParameter( "Autosave" )->toBool() ) {
		_autosavetimer->start( lastinterval * 60 * 1000 );
	}
}
void AbstractSpectrumModel::autoSave() {
	//qDebug() << "AbstractSpectrumModel::autoSave()";
	QString filename = _config->getParameter( "Autosave/Filename" )->toString();
	filename.replace( "DATE", QDateTime::currentDateTime().toString( "yyyyMMdd-hhmmss" ) );
	filename.replace( "MTIME", _mtime.toString( "yyyyMMdd-hhmmss" ) );
	saveToFile( filename );
}


Abstract1DSpectrumModel::Abstract1DSpectrumModel( QObject* p ) : AbstractSpectrumModel( p ) {
}
int Abstract1DSpectrumModel::columnCount( const QModelIndex& ) const {
	return 1;
}
QVariant Abstract1DSpectrumModel::headerData( int section, Qt::Orientation orientation, int role ) const {
	if ( role != Qt::DisplayRole )
		return QVariant();

	if ( orientation == Qt::Vertical )
		return rowHeader( section );

	if ( orientation == Qt::Horizontal ) {
		return "Count";
	}

	return QVariant();
}


TwoDimSpectrumModel::TwoDimSpectrumModel( QObject* p )
	: AbstractSpectrumModel( p )
{
	_cache.rows = _cache.cols = 0;
	new ModelTest( this, this );

	_config->addParameter( "X-Axis", QVariant::Invalid );
	NSCore::NamedParameterTreeItem* xaxis = _config->getParameter( "X-Axis" );
	xaxis->addParameter( "Minimum", -2.0 );
	xaxis->getParameter( "Minimum" )->setRange( -INFINITY, INFINITY );
	xaxis->addParameter( "Maximum", 8.0 );
	xaxis->getParameter( "Maximum" )->setRange( -INFINITY, INFINITY );
	xaxis->addParameter( "Channels", 100 );
	xaxis->getParameter( "Channels" )->setRange( 1, pow( 2,16 ) );
	connect( xaxis, SIGNAL( valueChanged() ), this, SLOT( clearSpectrum() ) );
	_config->addParameter( "Y-Axis", QVariant::Invalid );
	NSCore::NamedParameterTreeItem* yaxis = _config->getParameter( "Y-Axis" );
	yaxis->addParameter( "Minimum", -2.0 );
	yaxis->getParameter( "Minimum" )->setRange( -INFINITY, INFINITY );
	yaxis->addParameter( "Maximum", 8.0 );
	yaxis->getParameter( "Maximum" )->setRange( -INFINITY, INFINITY );
	yaxis->addParameter( "Channels", 100 );
	yaxis->getParameter( "Channels" )->setRange( 1, pow( 2,16 ) );
	connect( yaxis, SIGNAL( valueChanged() ), this, SLOT( clearSpectrum() ) );
	clearSpectrum();
}
int TwoDimSpectrumModel::columnCount( const QModelIndex& parent ) const {
	if ( parent.isValid() ) return 0;
	return _cache.cols;
}
int TwoDimSpectrumModel::rowCount( const QModelIndex& parent ) const {
	if ( parent.isValid() ) return 0;
	return _cache.rows;
}
QModelIndex TwoDimSpectrumModel::index( int row, int col, const QModelIndex& parent ) const {
	if ( parent.isValid() || row < 0 || col < 0 || row >= rowCount() || col >= columnCount() ) {
		return QModelIndex();
	}
	return createIndex( row, col );
}
void TwoDimSpectrumModel::data( double x, double y ) {
	//qDebug() << "TwoDimSpectrumModel::data(" << x << "," << y << ")";
	if ( isnan( x ) || isnan( y ) )
		return;

	if ( !( x < _cache.xmax && x > _cache.xmin ) )
		return;
	if ( !( y < _cache.ymax && y > _cache.ymin ) )
		return;

	int _x = int( ( x - _cache.xmin ) / _cache.xstep );
	int _y = int( ( y - _cache.ymin ) / _cache.ystep );
	_data[ _x ][ _y ] += 1;
	QMutexLocker lock( &_changed_index_lock );
	_changed_index.append( createIndex( _y,_x ) );
}
QVariant TwoDimSpectrumModel::data( const QModelIndex& index, int role ) const {
	if ( role != Qt::DisplayRole )
		return QVariant();

	//qDebug() << "TwoDimSpectrumModel::data(" << index << "," << role << ")";

	if ( index.row() >= _cache.rows )
		return QVariant();
	if ( index.column() >= _cache.cols )
		return QVariant();

	if ( role == Qt::DisplayRole ) {
		data_t::const_iterator x = _data.find( index.column() );
		if ( x != _data.end() ) {
			std::map<int,int>::const_iterator y = x->second.find( index.row() );
			if ( y != x->second.end() )
				return y->second;
		}
	}

	return QVariant();
}
QVariant TwoDimSpectrumModel::headerData( int section, Qt::Orientation orientation, int role ) const {
	if ( role != Qt::DisplayRole )
		return QVariant();

	if ( orientation == Qt::Horizontal && section < _cache.cols )
		return ( _cache.xmax - _cache.xmin ) / _cache.cols * section + _cache.xmin;
	if ( orientation == Qt::Vertical && section < _cache.rows )
		return ( _cache.ymax - _cache.ymin ) / _cache.rows * section + _cache.ymin;

	return QVariant();
}
void TwoDimSpectrumModel::saveToFile( QString path ) {
	//qDebug() << "TwoDimSpectrumModel::saveToFile( QString" << path << ")";
	double xmin = _cache.xmin;
	//double xmax = _cache.xmax;
	double ymin = _cache.ymin;
	//double ymax = _cache.ymax;
	double xstep = _cache.xstep;
	double ystep = _cache.ystep;
	QFile file( path );
	if ( file.open( QIODevice::WriteOnly ) ) {
		QTextStream stream( &file );

		stream << "ValueX ValueY Count" << endl;
		//foreach( int x, _data.keys() ) {
		for ( data_t::iterator it=_data.begin(); it!=_data.end(); ++it ) {
			int x = ( *it ).first;
			//foreach( int y, _data[ x ].keys() ) {
			for ( std::map<int,int>::iterator it2=( *it ).second.begin(); it2!=( *it ).second.end(); ++it2 ) {
				int y = ( *it2 ).first;
				//stream << xmin + xstep*x << " " << ymin + ystep*y << " " << _data[ x ][ y ] << endl;
				stream << xmin + xstep*x << " " << ymin + ystep*y << " " << ( *it2 ).second << endl;
			}
		}

		file.close();
	}
}
void TwoDimSpectrumModel::clearSpectrum() {
	qDebug() << "TwoDimSpectrumModel::clearSpectrum()";
	//beginRemoveRows( QModelIndex(), 0, __rows );
	/*QList<int> keysx = _data.keys();
	while ( ! _data.isEmpty() ) {
		_data.take( keysx.takeFirst() ).clear();
	}*/
	_data.clear();
	//endRemoveRows();
	_mtime = QDateTime::currentDateTime();
	//beginInsertRows( QModelIndex(), 0, __rows );
	//beginInsertColumns( QModelIndex(), 0, __cols );
	//endInsertRows();
	//endInsertColumns();
	emit dataChanged( index( 0,0, QModelIndex() ), index( rowCount()-1, columnCount()-1, QModelIndex() ) );
	updateCache();
}
void TwoDimSpectrumModel::updateCache() {
	_cache.cols = _config->getParameter( "X-Axis/Channels" )->toInt();
	_cache.xmin = _config->getParameter( "X-Axis/Minimum" )->toDouble();
	_cache.xmax = _config->getParameter( "X-Axis/Maximum" )->toDouble();
	_cache.xstep = ( _cache.xmax - _cache.xmin ) / _cache.cols;
	_cache.rows = _config->getParameter( "Y-Axis/Channels" )->toInt();
	_cache.ymin = _config->getParameter( "Y-Axis/Minimum" )->toDouble();
	_cache.ymax = _config->getParameter( "Y-Axis/Maximum" )->toDouble();
	_cache.ystep = ( _cache.ymax - _cache.ymin ) / _cache.rows;
}


SpectrumModel::SpectrumModel( QObject* p ) : Abstract1DSpectrumModel( p ), _accuracy( 0.01 ) {
	new ModelTest( this, this );

	_config->addParameter( "accuracy", 0.01 );
	_config->getParameter( "accuracy" )->setMin( 1E-5 );
	_config->getParameter( "accuracy" )->setMax( 1E4 );
	connect( _config, SIGNAL( valueChanged() ), this, SLOT( accuracyChanged() ) );
}
SpectrumModel::~SpectrumModel() {
}

int SpectrumModel::rowCount( const QModelIndex& ) const {
	return _data.size();
}

QModelIndex SpectrumModel::index( int row, int col, const QModelIndex& parent ) const {
	if ( parent.isValid() || row<0 || col<0 || row > rowCount() || col > 0 ) return QModelIndex();
	return createIndex( row, col );
}
QVariant SpectrumModel::rowHeader( int row ) const {
	if ( row>=0 && row<_data.size() )
		return _data.at( row ).key;
	return QVariant();
}

QVariant SpectrumModel::data( const QModelIndex& index, int role ) const {
	//qDebug() << "SpectrumModel::data( const QModelIndex& " << index << ", int " << role << ")";
	if ( index.isValid() && role == Qt::DisplayRole && index.column() < _data.size() ) {
		return _data.at( index.row() ).count;
	}
	return QVariant();
}

void SpectrumModel::data( double n ) {
	//qDebug() << "SpectrumModel::data( double " << n << " )";
	n = n - fmod( n, _accuracy );
	bool exists = false;
	for ( int i=0; i<_data.size(); ++i ) {
		if ( _data[ i ].key == n ) {
			//qDebug() << "n(" << n << ") exists (index=" << i << ")";
			exists = true;
			_data[ i ].count += 1;
			QModelIndex index0 = createIndex( i, 0 );
			QModelIndex index1 = createIndex( i, 1 );
			emit dataChanged( index0, index1 );
		}
	}
	if ( exists == false ) {
		//qDebug() << "n(" << n << ") doesn't exist creating a new entry";
		int i=0;
		for ( i=0; i<_data.size() && _data[ i ].key < n; ++i ) { }
		beginInsertRows( parent( QModelIndex() ), i, i );
		_data.insert( i, _Pair( n ) );
		endInsertRows();
		data( n );
	}
}

void SpectrumModel::saveToFile( QString path ) {
	QFile file( path );
	if ( file.open( QIODevice::WriteOnly ) ) {
		QTextStream stream( &file );

		stream << "Value" << " " << "Count" << endl;
		for ( int i=0; i<_data.size(); ++i )
			stream << _data[ i ].key << " " << _data[ i ].count << endl;

		file.close();
	}
}

bool SpectrumModel::removeRows( int row, int count, const QModelIndex& parent ) {
	if ( row+count <= _data.size() ) {
		beginRemoveRows( parent, row, row+count );
		for ( int i=0; i<count; ++i )
			_data.removeAt( row );
		endRemoveRows();
		return true;
	}
	return false;
}

void SpectrumModel::accuracyChanged() {
	_accuracy = config()[ "accuracy" ]->toDouble();
}

ClassicSpectrumModel::ClassicSpectrumModel( QObject* p )
	: Abstract1DSpectrumModel( p )
	, _data_lock( new QReadWriteLock() )
{
	new ModelTest( this, this );

	//qDebug() << "ClassicSpectrumModel::ClassicSpectrumModel(" << p << ")";
	_config->addParameter( "Minimum", -10.0 );
	_config->getParameter( "Minimum" )->setRange( -INFINITY, INFINITY );
	_config->addParameter( "Maximum", 10.0 );
	_config->getParameter( "Maximum" )->setRange( -INFINITY, INFINITY );
	_config->addParameter( "Channels", 200 );
	_config->getParameter( "Channels" )->setRange( 1, pow( 2, 16 ) );
	_config->addParameter( "Picoseconds per Sample", 250 );
	_config->getParameter( "Picoseconds per Sample" )->setRange( 125, 5000 );
	_config->addParameter( "Key value for *.dat-file", "300" );
	connect( _config->getParameter( "Minimum" ), SIGNAL( valueChanged() ), this, SLOT( redoChannels() ) );
	connect( _config->getParameter( "Maximum" ), SIGNAL( valueChanged() ), this, SLOT( redoChannels() ) );
	connect( _config->getParameter( "Channels" ), SIGNAL( valueChanged() ), this, SLOT( redoChannels() ) );

	//redoChannels();
	clearSpectrum();
}
ClassicSpectrumModel::~ClassicSpectrumModel() {
	qDebug() << "ClassicSpectrumModel::~ClassicSpectrumModel()";
}

QModelIndex ClassicSpectrumModel::index( int row, int column, const QModelIndex& parent ) const {
	if ( parent.isValid() || row<0 || column<0 ) {
		return QModelIndex();
	}
	if ( row < rowCount() && column < columnCount() ) {
		return createIndex( row, column );
	}
	return QModelIndex();
}
int ClassicSpectrumModel::rowCount( const QModelIndex& parent ) const {
	if ( parent.isValid() ) {
		return 0;
	}
	//return _config->getParameter( "Channels" )->toInt();
	return _data.size();
}

void ClassicSpectrumModel::data( double n ) {
	//qDebug() << "ClassicSpectrumModel::data(" << n << ")";
	int rowc = rowCount();
	for ( int i=0; i<rowc-1; ++i ) {
		QReadLocker lock( _data_lock );
		if ( _data[ i ].key <= n && _data[ i+1 ].key > n ) {
			_data[ i ].count += 1;
			QMutexLocker lock2( &_changed_index_lock );
			_changed_index.push_back( createIndex( i, 1 ) );
			_goodpulses++;
		}
	}
}

QVariant ClassicSpectrumModel::data( const QModelIndex& index, int role ) const {
	//qDebug() << "ClassicSpectrumModel::data( const QModelIndex& " << index << ", int " << role << ")";
	QReadLocker lock( _data_lock );
	if ( index.isValid() && role == Qt::DisplayRole && index.row() < _data.size() ) {
		return _data.at( index.row() ).count;
	}
	return QVariant();
}
QVariant ClassicSpectrumModel::rowHeader( int row ) const {
	if ( row < _data.size() )
		return _data.at( row ).key;
	return QVariant();
}
void ClassicSpectrumModel::saveToFile( QString path ) {
	qDebug() << "ClassicSpectrumModel::saveToFile(" << path << ")";
	QReadLocker lock( _data_lock );
	QFile file( path );
	if ( file.open( QIODevice::WriteOnly ) ) {
		QTextStream stream( &file );

		if ( path.endsWith( ".dat" ) ) {
			stream << QString( "; Filename is '%1', KeyValue is '%2'\r\n" ).arg( path.section( "/", -1 ) ).arg( _config->getParameter( "Key value for *.dat-file" )->toString() );
			// First is bin-width
			stream << _config->getParameter( "Picoseconds per Sample" )->toDouble() * fabs( _data[ 0 ].key - _data[ 1 ].key ) / 1000 << "\r\n";
			// Second an arificial key value to identify the measurement.
			stream << _config->getParameter( "Key value for *.dat-file" )->toString() << "\r\n";
			// Third is the accuracy of the measurement
			stream << 0.200 << "\r\n";

			for ( int i=0; i<_data.size(); ++i )
				stream << " " << _data[ i ].count << "\r\n";
		}
		else if ( path.endsWith( ".csv" ) ) {
			for ( int i=0; i<_data.size(); ++i )
				stream << _data[ i ].key << " " << _data[ i ].count << endl;
		}

		file.close();
	}
}
void ClassicSpectrumModel::redoChannels() {
	//qDebug() << "Something changed, redoing channels";

	beginRemoveRows( parent( QModelIndex() ), 0, _data.size()-1 );
	{
		QWriteLocker lock( _data_lock );
		while ( _data.size() > 0 )
			_data.pop_back();
	}
	endRemoveRows();

	int channels = config()["Channels"]->toInt();
	double min = config()[ "Minimum" ]->toDouble();
	double max = config()[ "Maximum" ]->toDouble();
	double step = ( max - min ) / channels;

	beginInsertRows( parent( QModelIndex() ), 0, channels-1 );
	{
		QWriteLocker lock( _data_lock );
		for ( int i=0; i<channels; ++i )
			_data.push_back( _Pair( min + step * i ) );
	}
	endInsertRows();
}

