/*
    Copyright 2005-2008 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef SPECTRUMMODEL_H
#define SPECTRUMMODEL_H

#include <QtCore/QAbstractItemModel>
#include <QtCore/QMap>
#include <QtCore/QList>
#include <QtCore/QMutex>
#include <QtCore/QReadWriteLock>
#include <QtCore/QTimer>
#include <QtCore/QDateTime>

#include <namedparametertreeitem.h>

/**
 * @brief Abstract baseclass for SpectrumModels
 *
 * This class provides the basic interface for SpectrumModels. That way several
 * models can benefit from one gui.
 *
 * @see SpectrumGui
 */
class AbstractSpectrumModel : public QAbstractItemModel
{
	Q_OBJECT
	Q_PROPERTY( int goodpulses READ goodpulses )
	public:
		/// simple default-constructor
		AbstractSpectrumModel( QObject* p );

		// QAbstractItemModel::virtual functions
		QModelIndex parent( const QModelIndex& ) const;

		/// Access the configuration parameters
		NSCore::NamedParameterTreeItemPointer config() const { return NSCore::NamedParameterTreeItemPointer( _config ); }

		/**
		 * @brief return the supported filters
		 *
		 * Implement this to return the valid file-endings your model can save.
		 * Is used in the file selection dialog.
		 */
		virtual QString supportedFilters() const =0;

		int goodpulses() const { return _goodpulses; }

		virtual int dimensions() const { return 1; }

	public slots:
		/**
		 * @brief Save the spectrum to a file
		 */
		virtual void saveToFile( QString ) =0;

		/**
		 * @brief Clear the whole spectrum
		 */
		virtual void clearSpectrum() =0;

	private slots:

		void changedIndexTimerTimeout();

		void autoSaveChanged();
		void autoSave();

	protected:

		NSCore::NamedParameterTreeItem* _config;
		int _goodpulses;

		QTimer* _changed_index_timer;
		QMutex _changed_index_lock;
		QList<QModelIndex> _changed_index;

		QTimer* _autosavetimer;
		QDateTime _mtime;
};

class Abstract1DSpectrumModel : public AbstractSpectrumModel
{
	Q_OBJECT
	public:
		Abstract1DSpectrumModel( QObject* );

		QVariant headerData( int section, Qt::Orientation orientation, int role = Qt::DisplayRole ) const;
		int columnCount( const QModelIndex& = QModelIndex() ) const;

	public slots:
		/**
		 * @brief Adds data to the model
		 */
		virtual void data( double ) =0;

	protected:
		struct _Pair {
			double key;
			int count;
			_Pair( double k ) : key( k ), count( 0 ) {
			}
		};

		virtual QVariant rowHeader( int row ) const =0;

};

/**
 * @todo docs!
 */
class TwoDimSpectrumModel : public AbstractSpectrumModel
{
	Q_OBJECT
	public:
		TwoDimSpectrumModel( QObject* );

		QModelIndex index( int, int, const QModelIndex& ) const;
		int columnCount( const QModelIndex& = QModelIndex() ) const;
		int rowCount( const QModelIndex& = QModelIndex() ) const;
		QVariant headerData( int section, Qt::Orientation orientation, int role = Qt::DisplayRole ) const;
		QVariant data( const QModelIndex&, int = Qt::DisplayRole ) const;

		QString supportedFilters() const { return "Comma separated file (*.csv)"; }

		int dimensions() const { return 2; }

	public slots:
		/**
		 * @brief Adds data to the model
		 */
		void data( double, double );

		void saveToFile( QString );
		void clearSpectrum();

	private slots:
		void updateCache();

	private:
		typedef std::map< int, std::map<int, int> > data_t;
		data_t _data;

		/// Cache for clearSpectrum
		struct {
			int rows, cols;
			double xmin, xmax, xstep;
			double ymin, ymax, ystep;
		} _cache;
};

/**
 * @todo docs!
 */
class SpectrumModel : public Abstract1DSpectrumModel
{
	Q_OBJECT
	public:
		SpectrumModel( QObject* );
		virtual ~SpectrumModel();

		// QAbstractItemModel::virtual functions
		int rowCount( const QModelIndex& = QModelIndex() ) const;
		QVariant data( const QModelIndex&, int = Qt::DisplayRole ) const;
		QModelIndex index( int, int, const QModelIndex& ) const;

		QVariant rowHeader( int ) const;

		bool removeRows( int, int, const QModelIndex& = QModelIndex() );

		QString supportedFilters() const { return "Comma separated File (*.csv)"; }
	public slots:
		// Own access-functions
		void data( double );

		/// Save the data of the model into the named file.
		void saveToFile( QString );

		/// Clear the whole spectrum
		void clearSpectrum() { removeRows( 0, rowCount() ); }
	private slots:
		void accuracyChanged();
	private:
		QList <_Pair> _data;
		double _accuracy;
};

/**
 * @brief A model for classic Spectra
 *
 * This class provides a model for classic spectra which have a min and max and
 * a number of channels.
 *
 * @todo more docs
 */
class ClassicSpectrumModel : public Abstract1DSpectrumModel
{
	Q_OBJECT
	public:
		ClassicSpectrumModel( QObject* );
		~ClassicSpectrumModel();

		// QAbstractItemModel::virtual functions
		QModelIndex index( int, int, const QModelIndex& ) const;
		int rowCount( const QModelIndex& = QModelIndex() ) const;
		QVariant data( const QModelIndex&, int = Qt::DisplayRole ) const;

		QString supportedFilters() const { return "Comma separated File (*.csv)\nDAT for LT (*.dat)"; }

		QVariant rowHeader( int ) const;

	public slots:
		void data( double );

		void saveToFile( QString );

		void clearSpectrum() { redoChannels(); _goodpulses = 0; _mtime = QDateTime::currentDateTime(); }

	private slots:
		void redoChannels();

	private:
		QReadWriteLock* _data_lock;
		QList<_Pair> _data;

};

#endif // SPECTRUMMODEL_H

